from abc import abstractmethod

from ecdsa.der import remove_sequence, UnexpectedDER, remove_integer

from .address import Address
from .constants import BECH32_ADDR_ACC_PREFIX
from .private_key import PrivateKey
from .public_key import PublicKey
from ..cosmos_app import CosmosApp
from ..utils import bip44_to_list


class Signer:
    @abstractmethod
    def get_public_key(self) -> PublicKey:
        """Gets public key associated with the signer.

        Returns:
            A PublicKey instance.
        """

        pass

    @abstractmethod
    def get_address(self) -> Address:
        """Gets address associated with the signer.

        Returns:
            An Address instance.
        """

        pass

    @abstractmethod
    def sign(self, msg: bytes) -> bytes:
        """Signs a message with the signer.

        Args:
            msg: The message to sign in bytes.

        Returns:
            Signature in bytes.
        """

        pass


class PrivateKeySigner(Signer):
    def __init__(self, private_key: PrivateKey):
        self.private_key = private_key

    def get_public_key(self) -> PublicKey:
        return self.private_key.to_public_key()

    def get_address(self) -> Address:
        return self.private_key.to_public_key().to_address()

    def sign(self, msg: bytes) -> bytes:
        return self.private_key.sign(bytes(msg))


class LedgerSigner(Signer):
    def __init__(self, path: str, app: CosmosApp):
        self.cosmos_app = app if app is not None else CosmosApp(bip44_to_list(path))

    def get_public_key(self) -> PublicKey:
        return PublicKey.from_hex(self.cosmos_app.ins_get_addr_secp256k1(BECH32_ADDR_ACC_PREFIX, False).public_key)

    def get_address(self) -> Address:
        return Address.from_acc_bech32(
            self.cosmos_app.ins_get_addr_secp256k1(BECH32_ADDR_ACC_PREFIX, False).address.decode()
        )

    def sign(self, msg: bytes) -> bytes:
        data, remaining_data = remove_sequence(self.cosmos_app.sign_secp256k1(msg))
        if remaining_data:
            raise UnexpectedDER("Unexpected remainder")

        r, remaining_content = remove_integer(data)
        s, remaining_content = remove_integer(remaining_content)
        if remaining_content:
            raise UnexpectedDER("Unexpected remainder")

        return r.to_bytes(32, "big") + s.to_bytes(32, "big")



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: cosmos/capability/v1beta1/capability.proto, cosmos/capability/v1beta1/genesis.proto
# plugin: python-betterproto
from dataclasses import dataclass
from typing import List

import betterproto


@dataclass(eq=False, repr=False)
class Capability(betterproto.Message):
    """
    Capability defines an implementation of an object capability. The index
    provided to a Capability must be globally unique.
    """

    index: int = betterproto.uint64_field(1)


@dataclass(eq=False, repr=False)
class Owner(betterproto.Message):
    """
    Owner defines a single capability owner. An owner is defined by the name of
    capability and the module name.
    """

    module: str = betterproto.string_field(1)
    name: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class CapabilityOwners(betterproto.Message):
    """
    CapabilityOwners defines a set of owners of a single Capability. The set of
    owners must be unique.
    """

    owners: List["Owner"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class GenesisOwners(betterproto.Message):
    """
    GenesisOwners defines the capability owners with their corresponding index.
    """

    index: int = betterproto.uint64_field(1)
    """index is the index of the capability owner."""

    index_owners: "CapabilityOwners" = betterproto.message_field(2)
    """index_owners are the owners at the given index."""


@dataclass(eq=False, repr=False)
class GenesisState(betterproto.Message):
    """GenesisState defines the capability module's genesis state."""

    index: int = betterproto.uint64_field(1)
    """index is the capability global index."""

    owners: List["GenesisOwners"] = betterproto.message_field(2)
    """
    owners represents a map from index to owners of the capability index index
    key is string to allow amino marshalling.
    """



import time
from typing import List

from grpclib.client import Channel

from .data import ReferencePrice, ReferencePriceUpdated
from .exceptions import NotFoundError, EmptyMsgError
from .proto.cosmos.auth.v1beta1 import BaseAccount, QueryAccountRequest
from .proto.cosmos.auth.v1beta1 import QueryStub as AuthQueryStub
from .proto.cosmos.base.abci.v1beta1 import TxResponse
from .proto.cosmos.base.tendermint.v1beta1 import GetLatestBlockRequest, GetLatestBlockResponse
from .proto.cosmos.base.tendermint.v1beta1 import ServiceStub as TendermintServiceStub
from .proto.cosmos.crypto.secp256k1 import PubKey
from .proto.cosmos.tx.v1beta1 import GetTxRequest, BroadcastTxRequest, BroadcastMode, SimulateRequest, SimulateResponse
from .proto.cosmos.tx.v1beta1 import ServiceStub as TxServiceStub
from .proto.oracle.v1 import (
    DataSource,
    OracleScript,
    QueryDataSourceRequest,
    QueryOracleScriptRequest,
    QueryRequestRequest,
    QueryRequestResponse,
    QueryReportersRequest,
    QueryRequestPriceRequest,
    QueryRequestSearchRequest,
    QueryRequestSearchResponse,
)
from .proto.oracle.v1 import MsgStub as OracleMsgStub
from .proto.oracle.v1 import QueryStub as OracleQueryStub


class Client:
    """Class for instantiating a client side connection with BandChain."""

    def __init__(self, channel: Channel):
        self.__channel = channel
        self.stub_oracle = OracleQueryStub(self.__channel)
        self.stub_cosmos_tendermint = TendermintServiceStub(self.__channel)
        self.stub_auth = AuthQueryStub(self.__channel)
        self.stub_tx = TxServiceStub(self.__channel)
        self.stub_oracle_tx = OracleMsgStub(self.__channel)

    def __del__(self) -> None:
        self.close()

    def close(self) -> None:
        """Closes the connection."""

        self.__channel.close()

    @classmethod
    def from_endpoint(cls, grpc_endpoint: str, port: int, ssl: bool = True):
        """Creates a client instance from a given endpoint and port.

        Args:
            grpc_endpoint: The GRPC endpoint to connect too.
            port: The endpoint port.
            ssl: If true, SSL context is used.

        Returns:
            A Client instance.
        """

        return cls(Channel(host=grpc_endpoint, port=port, ssl=ssl))

    async def get_data_source(self, id: int) -> DataSource:
        """Gets a data source's details.

        Args:
            id: The data source ID.

        Returns:
            The data source details.
        """

        resp = await self.stub_oracle.data_source(QueryDataSourceRequest(data_source_id=id))
        return resp.data_source

    async def get_oracle_script(self, id: int) -> OracleScript:
        """Gets an oracle script's details.

        Args:
            id: The oracle script ID.

        Returns:
            The oracle script details.
        """

        resp = await self.stub_oracle.oracle_script(QueryOracleScriptRequest(oracle_script_id=id))
        return resp.oracle_script

    async def get_request_by_id(self, id: int) -> QueryRequestResponse:
        """Gets a request's details.

        Args:
            id: The request ID.

        Returns:
            The request details.
        """

        resp = await self.stub_oracle.request(QueryRequestRequest(request_id=id))
        return resp

    async def get_reporters(self, validator: str) -> List[str]:
        """Gets a list of reporters associated with a given validator.

        Args:
            validator: The validator's address.

        Returns:
            A list of reporter addresses.
        """

        resp = await self.stub_oracle.reporters(QueryReportersRequest(validator_address=validator))
        return resp.reporter

    async def get_latest_block(self) -> GetLatestBlockResponse:
        """Gets the latest block on chain.

        Returns:
            The details of the latest block.
        """

        return await self.stub_cosmos_tendermint.get_latest_block(GetLatestBlockRequest())

    async def get_account(self, address: str) -> BaseAccount:
        """Gets the account details of a specified address.

        Args:
            address: The address to retrieve the detail of.

        Returns:
            The account details.
        """

        try:
            resp = await self.stub_auth.account(QueryAccountRequest(address=address))
            account = BaseAccount()
            pub_key = PubKey()

            account.parse(resp.account.value)
            account.pub_key = pub_key.parse(account.pub_key.value)
        except Exception as e:
            raise e

        return account

    async def get_request_id_by_tx_hash(self, tx_hash: str) -> List[int]:
        """Gets the request ID of a given transaction hash.

        Args:
            tx_hash: The transaction hash to retrieve the request ID of.

        Returns:
            A list of request IDs.
        """

        tx = await self.stub_tx.get_tx(GetTxRequest(hash=tx_hash))
        request_ids = []
        for log in tx.tx_response.logs:
            request_event = [event for event in log.events if event.type == "request" or event.type == "report"]
            if len(request_event) == 1:
                attrs = request_event[0].attributes
                attr_id = [attr for attr in attrs if attr.key == "id"]
                if len(attr_id) == 1:
                    request_id = attr_id[0].value
                    request_ids.append(int(request_id))
        if len(request_ids) == 0:
            raise NotFoundError("Request Id is not found")
        return request_ids

    async def send_tx_sync_mode(self, tx_bytes: bytes) -> TxResponse:
        """Sends a transaction in sync mode.

        Sends a transaction and waits until the transaction has passed the CheckTx phase.

        Args:
            tx_bytes: A signed transaction in raw bytes.

        Returns:
            The transaction response.
        """

        resp = await self.stub_tx.broadcast_tx(
            BroadcastTxRequest(tx_bytes=tx_bytes, mode=BroadcastMode.BROADCAST_MODE_SYNC)
        )
        return resp.tx_response

    async def send_tx_async_mode(self, tx_bytes: bytes) -> TxResponse:
        """Sends a transaction in async mode.

        Sends a transaction and returns the response immediately, without waiting for the transaction process.

        Args:
            tx_bytes: A signed transaction in raw bytes.

        Returns:
            The transaction response.
        """

        resp = await self.stub_tx.broadcast_tx(
            BroadcastTxRequest(tx_bytes=tx_bytes, mode=BroadcastMode.BROADCAST_MODE_ASYNC)
        )
        return resp.tx_response

    async def send_tx_block_mode(self, tx_bytes: bytes) -> TxResponse:
        """Sends a transaction in block mode.

        Sends a transaction and waits until the transaction has been committed to a block before returning the response.

        Args:
            tx_bytes: A signed transaction in raw bytes.

        Returns:
            The transaction response.
        """

        resp = await self.stub_tx.broadcast_tx(
            BroadcastTxRequest(tx_bytes=tx_bytes, mode=BroadcastMode.BROADCAST_MODE_BLOCK)
        )
        return resp.tx_response

    async def get_chain_id(self) -> str:
        """Gets the chain ID.

        Returns:
            The chain ID.
        """

        latest_block = await self.get_latest_block()
        return latest_block.block.header.chain_id

    async def get_reference_data(self, pairs: List[str], min_count: int, ask_count: int) -> List[ReferencePrice]:
        """Gets the rates of the given cryptocurrency pairs.

        Args:
            pairs: A list of cryptocurrency pairs.
            min_count: The minimum number of validators necessary for the request to proceed to the execution phase.
            ask_count: The number of validators requested to respond to the corresponding request.

        Returns:
            A list of reference prices, which contains the pair, its rate and its last updated time.
        """

        if len(pairs) == 0:
            raise EmptyMsgError("Pairs are required")

        symbols = set([symbol for pair in pairs for symbol in pair.split("/") if symbol != "USD"])

        price_data = await self.stub_oracle.request_price(
            QueryRequestPriceRequest(
                symbols=list(symbols),
                min_count=min_count,
                ask_count=ask_count,
            )
        )

        symbol_dict = {
            "USD": {
                "multiplier": 1000000000,
                "px": 1000000000,
                "resolve_time": int(time.time()),
            }
        }

        for price in price_data.price_results:
            symbol_dict[price.symbol] = {
                "multiplier": price.multiplier,
                "px": price.px,
                "resolve_time": price.resolve_time,
            }

        results = []
        for pair in pairs:
            base_symbol, quote_symbol = pair.split("/")

            quote_rate = int(symbol_dict[base_symbol]["px"]) / int(symbol_dict[base_symbol]["multiplier"])
            base_rate = int(symbol_dict[quote_symbol]["px"]) / int(symbol_dict[quote_symbol]["multiplier"])
            rate = quote_rate / base_rate

            rate_updated_at = ReferencePriceUpdated(
                int(symbol_dict[base_symbol]["resolve_time"]),
                int(symbol_dict[quote_symbol]["resolve_time"]),
            )

            results.append(ReferencePrice(pair, rate=rate, updated_at=rate_updated_at))

        return results

    async def get_latest_request(
        self, oid: int, calldata: str, min_count: int, ask_count: int
    ) -> QueryRequestSearchResponse:
        """Gets the latest on chain request.

        Args:
            oid: The oracle script ID.
            calldata: The calldata of the request.
            min_count: The minimum number of validators necessary for the request to proceed to the execution phase.
            ask_count: The number of validators requested to respond to the corresponding request.

        Returns:
            The request details.
        """
        return await self.stub_oracle.request_search(
            QueryRequestSearchRequest(
                oracle_script_id=oid, calldata=calldata, ask_count=ask_count, min_count=min_count
            )
        )

    async def get_tx_response(self, tx_hash: str) -> TxResponse:
        """Gets the tx response from a tx hash.

        Args:
            tx_hash: The transaction hash to retrieve the tx response of.

        Returns:
            The tx response.
        """
        resp = await self.stub_tx.get_tx(GetTxRequest(hash=tx_hash))
        return resp.tx_response

    async def simulate_tx(self, tx_bytes: bytes) -> SimulateResponse:
        """Simulates a transaction from the tx_bytes.

        Args:
            tx_bytes: A signed transaction in raw bytes.

        Returns:
            The simulated response.
        """
        return await self.stub_tx.simulate(SimulateRequest(tx_bytes=tx_bytes))



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: tendermint/crypto/keys.proto, tendermint/crypto/proof.proto
# plugin: python-betterproto
from dataclasses import dataclass
from typing import List

import betterproto


@dataclass(eq=False, repr=False)
class Proof(betterproto.Message):
    total: int = betterproto.int64_field(1)
    index: int = betterproto.int64_field(2)
    leaf_hash: bytes = betterproto.bytes_field(3)
    aunts: List[bytes] = betterproto.bytes_field(4)


@dataclass(eq=False, repr=False)
class ValueOp(betterproto.Message):
    key: bytes = betterproto.bytes_field(1)
    """Encoded in ProofOp.Key."""

    proof: "Proof" = betterproto.message_field(2)
    """To encode in ProofOp.Data"""


@dataclass(eq=False, repr=False)
class DominoOp(betterproto.Message):
    key: str = betterproto.string_field(1)
    input: str = betterproto.string_field(2)
    output: str = betterproto.string_field(3)


@dataclass(eq=False, repr=False)
class ProofOp(betterproto.Message):
    """
    ProofOp defines an operation used for calculating Merkle root The data
    could be arbitrary format, providing nessecary data for example
    neighbouring node hash
    """

    type: str = betterproto.string_field(1)
    key: bytes = betterproto.bytes_field(2)
    data: bytes = betterproto.bytes_field(3)


@dataclass(eq=False, repr=False)
class ProofOps(betterproto.Message):
    """ProofOps is Merkle proof defined by the list of ProofOps"""

    ops: List["ProofOp"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class PublicKey(betterproto.Message):
    """
    PublicKey defines the keys available for use with Tendermint Validators
    """

    ed25519: bytes = betterproto.bytes_field(1, group="sum")
    secp256_k1: bytes = betterproto.bytes_field(2, group="sum")



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: tendermint/types/block.proto, tendermint/types/evidence.proto, tendermint/types/params.proto, tendermint/types/types.proto, tendermint/types/validator.proto
# plugin: python-betterproto
import builtins
from dataclasses import dataclass
from datetime import (
    datetime,
    timedelta,
)
from typing import List

import betterproto

from .. import (
    crypto as _crypto__,
    version as _version__,
)


class BlockIdFlag(betterproto.Enum):
    """BlockIdFlag indicates which BlockID the signature is for"""

    BLOCK_ID_FLAG_UNKNOWN = 0
    BLOCK_ID_FLAG_ABSENT = 1
    BLOCK_ID_FLAG_COMMIT = 2
    BLOCK_ID_FLAG_NIL = 3


class SignedMsgType(betterproto.Enum):
    """SignedMsgType is a type of signed message in the consensus."""

    SIGNED_MSG_TYPE_UNKNOWN = 0
    SIGNED_MSG_TYPE_PREVOTE = 1
    """Votes"""

    SIGNED_MSG_TYPE_PRECOMMIT = 2
    SIGNED_MSG_TYPE_PROPOSAL = 32
    """Proposals"""


@dataclass(eq=False, repr=False)
class ValidatorSet(betterproto.Message):
    validators: List["Validator"] = betterproto.message_field(1)
    proposer: "Validator" = betterproto.message_field(2)
    total_voting_power: int = betterproto.int64_field(3)


@dataclass(eq=False, repr=False)
class Validator(betterproto.Message):
    address: bytes = betterproto.bytes_field(1)
    pub_key: "_crypto__.PublicKey" = betterproto.message_field(2)
    voting_power: int = betterproto.int64_field(3)
    proposer_priority: int = betterproto.int64_field(4)


@dataclass(eq=False, repr=False)
class SimpleValidator(betterproto.Message):
    pub_key: "_crypto__.PublicKey" = betterproto.message_field(1)
    voting_power: int = betterproto.int64_field(2)


@dataclass(eq=False, repr=False)
class PartSetHeader(betterproto.Message):
    """PartsetHeader"""

    total: int = betterproto.uint32_field(1)
    hash: bytes = betterproto.bytes_field(2)


@dataclass(eq=False, repr=False)
class Part(betterproto.Message):
    index: int = betterproto.uint32_field(1)
    bytes: builtins.bytes = betterproto.bytes_field(2)
    proof: "_crypto__.Proof" = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class BlockId(betterproto.Message):
    """BlockID"""

    hash: bytes = betterproto.bytes_field(1)
    part_set_header: "PartSetHeader" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class Header(betterproto.Message):
    """Header defines the structure of a Tendermint block header."""

    version: "_version__.Consensus" = betterproto.message_field(1)
    """basic block info"""

    chain_id: str = betterproto.string_field(2)
    height: int = betterproto.int64_field(3)
    time: datetime = betterproto.message_field(4)
    last_block_id: "BlockId" = betterproto.message_field(5)
    """prev block info"""

    last_commit_hash: bytes = betterproto.bytes_field(6)
    """hashes of block data"""

    data_hash: bytes = betterproto.bytes_field(7)
    validators_hash: bytes = betterproto.bytes_field(8)
    """hashes from the app output from the prev block"""

    next_validators_hash: bytes = betterproto.bytes_field(9)
    consensus_hash: bytes = betterproto.bytes_field(10)
    app_hash: bytes = betterproto.bytes_field(11)
    last_results_hash: bytes = betterproto.bytes_field(12)
    evidence_hash: bytes = betterproto.bytes_field(13)
    """consensus info"""

    proposer_address: bytes = betterproto.bytes_field(14)


@dataclass(eq=False, repr=False)
class Data(betterproto.Message):
    """Data contains the set of transactions included in the block"""

    txs: List[bytes] = betterproto.bytes_field(1)
    """
    Txs that will be applied by state @ block.Height+1. NOTE: not all txs here
    are valid.  We're just agreeing on the order first. This means that
    block.AppHash does not include these txs.
    """


@dataclass(eq=False, repr=False)
class Vote(betterproto.Message):
    """
    Vote represents a prevote, precommit, or commit vote from validators for
    consensus.
    """

    type: "SignedMsgType" = betterproto.enum_field(1)
    height: int = betterproto.int64_field(2)
    round: int = betterproto.int32_field(3)
    block_id: "BlockId" = betterproto.message_field(4)
    timestamp: datetime = betterproto.message_field(5)
    validator_address: bytes = betterproto.bytes_field(6)
    validator_index: int = betterproto.int32_field(7)
    signature: bytes = betterproto.bytes_field(8)


@dataclass(eq=False, repr=False)
class Commit(betterproto.Message):
    """
    Commit contains the evidence that a block was committed by a set of
    validators.
    """

    height: int = betterproto.int64_field(1)
    round: int = betterproto.int32_field(2)
    block_id: "BlockId" = betterproto.message_field(3)
    signatures: List["CommitSig"] = betterproto.message_field(4)


@dataclass(eq=False, repr=False)
class CommitSig(betterproto.Message):
    """CommitSig is a part of the Vote included in a Commit."""

    block_id_flag: "BlockIdFlag" = betterproto.enum_field(1)
    validator_address: bytes = betterproto.bytes_field(2)
    timestamp: datetime = betterproto.message_field(3)
    signature: bytes = betterproto.bytes_field(4)


@dataclass(eq=False, repr=False)
class Proposal(betterproto.Message):
    type: "SignedMsgType" = betterproto.enum_field(1)
    height: int = betterproto.int64_field(2)
    round: int = betterproto.int32_field(3)
    pol_round: int = betterproto.int32_field(4)
    block_id: "BlockId" = betterproto.message_field(5)
    timestamp: datetime = betterproto.message_field(6)
    signature: bytes = betterproto.bytes_field(7)


@dataclass(eq=False, repr=False)
class SignedHeader(betterproto.Message):
    header: "Header" = betterproto.message_field(1)
    commit: "Commit" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class LightBlock(betterproto.Message):
    signed_header: "SignedHeader" = betterproto.message_field(1)
    validator_set: "ValidatorSet" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class BlockMeta(betterproto.Message):
    block_id: "BlockId" = betterproto.message_field(1)
    block_size: int = betterproto.int64_field(2)
    header: "Header" = betterproto.message_field(3)
    num_txs: int = betterproto.int64_field(4)


@dataclass(eq=False, repr=False)
class TxProof(betterproto.Message):
    """
    TxProof represents a Merkle proof of the presence of a transaction in the
    Merkle tree.
    """

    root_hash: bytes = betterproto.bytes_field(1)
    data: bytes = betterproto.bytes_field(2)
    proof: "_crypto__.Proof" = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class ConsensusParams(betterproto.Message):
    """
    ConsensusParams contains consensus critical parameters that determine the
    validity of blocks.
    """

    block: "BlockParams" = betterproto.message_field(1)
    evidence: "EvidenceParams" = betterproto.message_field(2)
    validator: "ValidatorParams" = betterproto.message_field(3)
    version: "VersionParams" = betterproto.message_field(4)


@dataclass(eq=False, repr=False)
class BlockParams(betterproto.Message):
    """BlockParams contains limits on the block size."""

    max_bytes: int = betterproto.int64_field(1)
    """Max block size, in bytes. Note: must be greater than 0"""

    max_gas: int = betterproto.int64_field(2)
    """Max gas per block. Note: must be greater or equal to -1"""

    time_iota_ms: int = betterproto.int64_field(3)
    """
    Minimum time increment between consecutive blocks (in milliseconds) If the
    block header timestamp is ahead of the system clock, decrease this value.
    Not exposed to the application.
    """


@dataclass(eq=False, repr=False)
class EvidenceParams(betterproto.Message):
    """EvidenceParams determine how we handle evidence of malfeasance."""

    max_age_num_blocks: int = betterproto.int64_field(1)
    """
    Max age of evidence, in blocks. The basic formula for calculating this is:
    MaxAgeDuration / {average block time}.
    """

    max_age_duration: timedelta = betterproto.message_field(2)
    """
    Max age of evidence, in time. It should correspond with an app's "unbonding
    period" or other similar mechanism for handling [Nothing-At-Stake
    attacks](https://github.com/ethereum/wiki/wiki/Proof-of-Stake-FAQ#what-is-
    the-nothing-at-stake-problem-and-how-can-it-be-fixed).
    """

    max_bytes: int = betterproto.int64_field(3)
    """
    This sets the maximum size of total evidence in bytes that can be committed
    in a single block. and should fall comfortably under the max block bytes.
    Default is 1048576 or 1MB
    """


@dataclass(eq=False, repr=False)
class ValidatorParams(betterproto.Message):
    """
    ValidatorParams restrict the public key types validators can use. NOTE:
    uses ABCI pubkey naming, not Amino names.
    """

    pub_key_types: List[str] = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class VersionParams(betterproto.Message):
    """VersionParams contains the ABCI application version."""

    app_version: int = betterproto.uint64_field(1)


@dataclass(eq=False, repr=False)
class HashedParams(betterproto.Message):
    """
    HashedParams is a subset of ConsensusParams. It is hashed into the
    Header.ConsensusHash.
    """

    block_max_bytes: int = betterproto.int64_field(1)
    block_max_gas: int = betterproto.int64_field(2)


@dataclass(eq=False, repr=False)
class Evidence(betterproto.Message):
    duplicate_vote_evidence: "DuplicateVoteEvidence" = betterproto.message_field(1, group="sum")
    light_client_attack_evidence: "LightClientAttackEvidence" = betterproto.message_field(2, group="sum")


@dataclass(eq=False, repr=False)
class DuplicateVoteEvidence(betterproto.Message):
    """
    DuplicateVoteEvidence contains evidence of a validator signed two
    conflicting votes.
    """

    vote_a: "Vote" = betterproto.message_field(1)
    vote_b: "Vote" = betterproto.message_field(2)
    total_voting_power: int = betterproto.int64_field(3)
    validator_power: int = betterproto.int64_field(4)
    timestamp: datetime = betterproto.message_field(5)


@dataclass(eq=False, repr=False)
class LightClientAttackEvidence(betterproto.Message):
    """
    LightClientAttackEvidence contains evidence of a set of validators
    attempting to mislead a light client.
    """

    conflicting_block: "LightBlock" = betterproto.message_field(1)
    common_height: int = betterproto.int64_field(2)
    byzantine_validators: List["Validator"] = betterproto.message_field(3)
    total_voting_power: int = betterproto.int64_field(4)
    timestamp: datetime = betterproto.message_field(5)


@dataclass(eq=False, repr=False)
class EvidenceList(betterproto.Message):
    evidence: List["Evidence"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class Block(betterproto.Message):
    header: "Header" = betterproto.message_field(1)
    data: "Data" = betterproto.message_field(2)
    evidence: "EvidenceList" = betterproto.message_field(3)
    last_commit: "Commit" = betterproto.message_field(4)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: cosmos/base/abci/v1beta1/abci.proto
# plugin: python-betterproto
from dataclasses import dataclass
from typing import List

import betterproto
import betterproto.lib.google.protobuf as betterproto_lib_google_protobuf

from .....tendermint import abci as ____tendermint_abci__


@dataclass(eq=False, repr=False)
class TxResponse(betterproto.Message):
    """
    TxResponse defines a structure containing relevant tx data and metadata.
    The tags are stringified and the log is JSON decoded.
    """

    height: int = betterproto.int64_field(1)
    """The block height"""

    txhash: str = betterproto.string_field(2)
    """The transaction hash."""

    codespace: str = betterproto.string_field(3)
    """Namespace for the Code"""

    code: int = betterproto.uint32_field(4)
    """Response code."""

    data: str = betterproto.string_field(5)
    """Result bytes, if any."""

    raw_log: str = betterproto.string_field(6)
    """
    The output of the application's logger (raw string). May be non-
    deterministic.
    """

    logs: List["AbciMessageLog"] = betterproto.message_field(7)
    """
    The output of the application's logger (typed). May be non-deterministic.
    """

    info: str = betterproto.string_field(8)
    """Additional information. May be non-deterministic."""

    gas_wanted: int = betterproto.int64_field(9)
    """Amount of gas requested for transaction."""

    gas_used: int = betterproto.int64_field(10)
    """Amount of gas consumed by transaction."""

    tx: "betterproto_lib_google_protobuf.Any" = betterproto.message_field(11)
    """The request transaction bytes."""

    timestamp: str = betterproto.string_field(12)
    """
    Time of the previous block. For heights > 1, it's the weighted median of
    the timestamps of the valid votes in the block.LastCommit. For height == 1,
    it's genesis time.
    """

    events: List["____tendermint_abci__.Event"] = betterproto.message_field(13)
    """
    Events defines all the events emitted by processing a transaction. Note,
    these events include those emitted by processing all the messages and those
    emitted from the ante handler. Whereas Logs contains the events, with
    additional metadata, emitted only by processing the messages. Since:
    cosmos-sdk 0.42.11, 0.44.5, 0.45
    """


@dataclass(eq=False, repr=False)
class AbciMessageLog(betterproto.Message):
    """
    ABCIMessageLog defines a structure containing an indexed tx ABCI message
    log.
    """

    msg_index: int = betterproto.uint32_field(1)
    log: str = betterproto.string_field(2)
    events: List["StringEvent"] = betterproto.message_field(3)
    """
    Events contains a slice of Event objects that were emitted during some
    execution.
    """


@dataclass(eq=False, repr=False)
class StringEvent(betterproto.Message):
    """
    StringEvent defines en Event object wrapper where all the attributes
    contain key/value pairs that are strings instead of raw bytes.
    """

    type: str = betterproto.string_field(1)
    attributes: List["Attribute"] = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class Attribute(betterproto.Message):
    """
    Attribute defines an attribute wrapper where the key and value are strings
    instead of raw bytes.
    """

    key: str = betterproto.string_field(1)
    value: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class GasInfo(betterproto.Message):
    """GasInfo defines tx execution gas context."""

    gas_wanted: int = betterproto.uint64_field(1)
    """GasWanted is the maximum units of work we allow this tx to perform."""

    gas_used: int = betterproto.uint64_field(2)
    """GasUsed is the amount of gas actually consumed."""


@dataclass(eq=False, repr=False)
class Result(betterproto.Message):
    """Result is the union of ResponseFormat and ResponseCheckTx."""

    data: bytes = betterproto.bytes_field(1)
    """
    Data is any data returned from message or handler execution. It MUST be
    length prefixed in order to separate data from multiple message executions.
    """

    log: str = betterproto.string_field(2)
    """Log contains the log information from message or handler execution."""

    events: List["____tendermint_abci__.Event"] = betterproto.message_field(3)
    """
    Events contains a slice of Event objects that were emitted during message
    or handler execution.
    """


@dataclass(eq=False, repr=False)
class SimulationResponse(betterproto.Message):
    """
    SimulationResponse defines the response generated when a transaction is
    successfully simulated.
    """

    gas_info: "GasInfo" = betterproto.message_field(1)
    result: "Result" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class MsgData(betterproto.Message):
    """
    MsgData defines the data returned in a Result object during message
    execution.
    """

    msg_type: str = betterproto.string_field(1)
    data: bytes = betterproto.bytes_field(2)


@dataclass(eq=False, repr=False)
class TxMsgData(betterproto.Message):
    """
    TxMsgData defines a list of MsgData. A transaction will have a MsgData
    object for each message.
    """

    data: List["MsgData"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class SearchTxsResult(betterproto.Message):
    """SearchTxsResult defines a structure for querying txs pageable"""

    total_count: int = betterproto.uint64_field(1)
    """Count of all txs"""

    count: int = betterproto.uint64_field(2)
    """Count of txs in current page"""

    page_number: int = betterproto.uint64_field(3)
    """Index of current page, start from 1"""

    page_total: int = betterproto.uint64_field(4)
    """Count of total pages"""

    limit: int = betterproto.uint64_field(5)
    """Max count txs per page"""

    txs: List["TxResponse"] = betterproto.message_field(6)
    """List of txs in current page"""



import re
from typing import List

from .exceptions import NotBip44Error, IncorrectLengthError


def is_bip44(derivation_path: str) -> bool:
    """Checks if the given derivation path is a valid bip44 path.

    Args:
        derivation_path: Derivation path as a string

    Returns:
        Whether the path is valid or not
    """
    if re.search("^m(/[0-9]+['hH]?)*$", derivation_path) is not None:
        return True
    else:
        return False


def bip44_to_list(derivation_path: str) -> List[int]:
    """Converts the given string formatted derivation path into a list
    Args:
        derivation_path: Derivation path as a string

    Returns:
        Derivation path as a list

    Raises:
        NotBip44Error: If derivation path is not valid
    """
    try:
        return [int(x) for x in derivation_path.replace("'", "").split("/")[1:]]
    except ValueError:
        raise NotBip44Error()
    except Exception as e:
        raise e


def get_bip32_byte(derivation_path: List[int], harden_count: int) -> bytes:
    """Encodes BIP32 as a byte

    Args:
        derivation_path: The derivation path as a list
        harden_count: The depth level of keys to harden

    Returns:
        BIP32 as a byte
    """
    if len(derivation_path) != 5:
        raise IncorrectLengthError("Incorrect path length")

    path_bytes = b""
    for i, idx in enumerate(derivation_path):
        if i < harden_count:
            idx = 0x80000000 | idx
        path_bytes += idx.to_bytes(4, "little")
    return path_bytes


def split_packet(byte_message: bytes) -> List[bytes]:
    """
    Splits a message into packets of size 250 bytes each

    Args:
        byte_message: Message as in bytes

    Returns:
        Message split into packets
    """
    return [bytes(byte_message[i : i + 250]) for i in range(0, len(byte_message), 250)]
